/*
 *  prommer - EPROM burner for parallel-port EPROMmer.
 *
 *  Sat Feb 22 06:20:22 EET 1997, junki
 *
 *  A bidirectional parallel-port is required.
 *  Databits are connected to 74245, 72374 and 74373.
 *  Latch pulse comes from INIT.
 *  245 /ENB and prom /CE comes from STROBE.
 *  245 DIR and /OE comes from SELECT.
 *  A relay is switched by FEED.
 *
 *  74245 should be oriented so that DIR high buffers data
 *  FROM parallel port INTO the prom.
 */

#include <machine/busio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <signal.h>
#include <setjmp.h>

#define IDLE     0x00 /* all inactive */

#define ADDRESS  0x04 /* INIT */
#define ENB      0x01 /* STROBE */
#define DIR      0x28 /* SELECT and LPTDIR */
#define VPP      0x02 /* FEED, switch relay on */

int burn_time  = 1000; /* usec */
int delay_time = 5;    /* loops */

int data    = 0x3bc;
int control = 0x3be;

int ctl;

jmp_buf jbuf;
unsigned errors;
unsigned from  = 0;
unsigned to    = 0x7FFF;

int offset, woff, roff;

int Rflag = 0;
int Sflag = 0;
int Vflag = 0;
int Qflag = 0;
int Oflag = 0;
int Eflag = 0;

enum {
	DEFLT,
	DUMP,
	BURN,
	VERIFY,
	BURN_VERIFY_BURN,
	EEUNLOCK,
	BLANKCHECK,
} mode;

int prom_type;
int image_size;
int do_bell;

u_char image[0x10000];
u_char cksok[0x10000];

getb(a) { return ((a < image_size) ? image[a] : -1); }

void
intr(int sig)
{
	exit(128 + sig);
}

void
skip_this(int sig)
{
	longjmp(jbuf, 1);
}

close_prommer()
{
	outb(control, ctl = IDLE);
	if (do_bell)
		write(2, "\a\0\a\a", 4);
}

open_prommer()
{
	outb(control, ctl = IDLE);

	set_addr(0);

	atexit(close_prommer);

	signal(SIGINT,  intr);
	signal(SIGTERM, intr);
	signal(SIGPIPE, intr);
}

main(argc, argv)
	char **argv;
{
	extern int optind, opterr;
	extern char *optarg;
	int c;

	if (!init_busio()) {
		perror("prommer: Cannot get bus IO privileges");
		exit(1);
	}
	setuid(getuid());

	opterr = 0;

	while ((c = getopt(argc, argv, "3:6:1:2:5:rwvxbVQRSOEe:d:D:f:t:o:")) != -1)
		switch (c) {
		default:
	baduse:
			fprintf(stderr, "\
Usage: prommer [-rwv][-16/32/64/128/256/512][-VQRS][-OE]\n\
               [-d rtim][-D btim][-f frm][-t to]\n\
               [-eeunlock]\n\
               [-b]\n\
               [file]\n");
			exit(2);
		case 'O':
			Oflag++;
			break;
		case 'E':
			Eflag++;
			break;
		case 'R':
			Rflag++;
			break;
		case 'S':
			Sflag++;
			break;
		case 'V':
			Vflag++;
			break;
		case 'Q':
			Qflag++;
			break;
		case 'd':
			delay_time = atoi(optarg);
			break;
		case 'D':
			burn_time = atoi(optarg);
			break;
		case 'f':
			sscanf(optarg, "%x", &from);
			break;
		case 't':
			sscanf(optarg, "%x", &to);
			break;
		case 'o':
			sscanf(optarg, "%x", &offset);
			break;
		case 'r':
			if (mode)
				goto baduse;
			mode = DUMP;
			break;

		case 'w':
			if (mode)
				goto baduse;
			mode = BURN;
			break;
		case 'v':
			if (mode)
				goto baduse;
			mode = VERIFY;
			break;
		case 'x':
			if (mode)
				goto baduse;
			mode = BURN_VERIFY_BURN;
			break;
		case 'b':
			if (mode)
				goto baduse;
			mode = BLANKCHECK;
			break;

		case '3':
			if (!strcmp(optarg, 1+"32")) {
				prom_type = 32;
				to = 0xFFF;
				roff = 0x6800;
				break;
			}
			goto baduse;
		case '6':
			if (!strcmp(optarg, 1+"64")) {
				prom_type = 64;
				to = 0x1FFF;
				roff = 0x4000;
				break;
			}
			goto baduse;
		case '1':
			if (!strcmp(optarg, 1+"16")) {
				prom_type = 16;
				to = 0x7FF;
				roff = 0x6800;
				break;
			}
			if (!strcmp(optarg, 1+"128")) {
				prom_type = 128;
				to = 0x3FFF;
				roff = 0x4000;
				break;
			}
			goto baduse;
		case '2':
			if (!strcmp(optarg, 1+"256")) {
				prom_type = 256;
				to = 0x7FFF;
				break;
			}
			goto baduse;
		case '5':
			if (!strcmp(optarg, 1+"512")) {
				prom_type = 512;
				to = 0xFFFF;
				break;
			}
			goto baduse;

		case 'e':
			if (!strcmp(optarg, 1+"eeunlock")) {
				mode = EEUNLOCK;
				break;
			}
		}

	if (mode == DEFLT)
		mode = DUMP;

	if (optind < argc && mode != DUMP) {
		if (freopen(argv[optind], "r", stdin) == NULL) {
			perror(argv[optind]);
			exit(1);
		}
		++optind;
	}
	if (optind < argc)
		goto baduse;

	image_size = to - from + 1;

	if (image_size > 0x10000) {
		fprintf(stderr, "Too large.\n");
		exit(1);
	}
	if (mode != DUMP && mode != EEUNLOCK && mode != BLANKCHECK) {
		image_size = fread(image, 1, image_size, stdin);
		if (image_size <= 0) {
			perror("read");
			exit(1);
		}
		to = from + image_size - 1;
	}
	open_prommer();

	switch (mode) {
	default:
		dumprom();
		break;
	case EEUNLOCK:
		eeunlock();
		break;
	case BURN:
		burn();
		break;
	case VERIFY:
		verify();
		if (Vflag)
			fprintf(stderr, "%d errors\n", errors);
		break;
	case BLANKCHECK:
		check();
		break;

	case BURN_VERIFY_BURN:
		setjmp(jbuf);
		signal(SIGINT, skip_this);
		fprintf(stderr, "\n");
		for (;;) {
			char buf[80];
			fprintf(stderr, "qfvbB ? ");
			if (read(2, buf, sizeof(buf)) <= 0 || *buf == 'q')
				break;
			switch (*buf) {
			case 'f':
				bzero(cksok, sizeof(cksok));
				break;
			case 'B':
				for (;;)
					burn();
				break;
			case 'b':
				burn();
				break;
			case 'v':
				verify();
				fprintf(stderr, "%d errors\n", errors);
				break;
			}
		}
		break;
	}
	exit(errors < 256 ? errors : 255);
}

verify()
{
	unsigned fc, rc, addr;
	char x = 0;

	errors = 0;
	for (addr = from; addr <= to; x ^= 1) {
		if ((fc = getb(addr)) == -1)
			break;
		if (Eflag && x)
			continue;
		if (Oflag && !x)
			continue;
		rc = read_byte(addr);
		if (rc != fc) {
			cksok[addr] = 0;
			if (++errors < 20 && !Qflag) {
				fprintf(stderr, "%04X eprom %02x, file %02x\n", addr, rc, fc);
			}
		} else {
			cksok[addr] = 1;
		}
		addr++;
	}
}

burn()
{
	unsigned c, addr;
	char x = 0;

	for (addr = from; addr <= to; x ^= 1) {
		if ((c = getb(addr)) == -1)
			break;
		if (Eflag && x)
			continue;
		if (Oflag && !x)
			continue;
		if (!cksok[addr])
			write_byte(addr, c);
		addr++;
	}
}

eeunlock()
{
	write_byte(0x1555, 0xAA);
	write_byte(0x0AAA, 0x55);
	write_byte(0x1555, 0x80);
	write_byte(0x1555, 0xAA);
	write_byte(0x0AAA, 0x55);
	write_byte(0x1555, 0x80);
	write_byte(0x0001, 0xAB);
}

check()
{
	unsigned addr, addr0, addrd;
	unsigned char c;

	for (addr = from; addr <= to; ++addr) {
		c = read_byte(addr);
		if (c != 0xFF) {
			if (!errors++)
				addr0 = addr;
			addrd = addr;
		}
	}
	if (errors)
		fprintf(stderr, "%d nonblank, first 0x%04x, last 0x%04x\n", errors, addr0, addrd);
	else if (Vflag)
		printf("blank\n");
}

dumprom()
{
	int i;
	unsigned addr;
	unsigned char c, bytes[16];

	if (!Rflag) {
		from &= ~0xF;
		to   &= ~0xF;
	}
	for (addr = from; addr <= to; ) {
		if (Rflag || Sflag) {
			c = read_byte(addr++);
			if (Sflag) {
				c &= 0x7F;
				if ((c >= ' ' && c < 0x7F) || c == '\n' || c == '\t') {
					putchar(c);
					fflush(stdout);
				}
			} else {
				putchar(c);
			}
			continue;
		}
		printf("%04x:", addr);

		for (i = 0; i < sizeof(bytes); ++i)
			bytes[i] = read_byte(addr + i);
		for (i = 0; i < sizeof(bytes); ++i)
			printf(" %02x", bytes[i]);

		printf("  ", addr);
		for (i = 0; i < sizeof(bytes); ++i)
			putchar((bytes[i] >= ' ' && bytes[i] < 0177) ? bytes[i] : '.');
		putchar('\n');
		addr += sizeof(bytes);

		fflush(stdout);
	}
}

burn_delay()
{
	if (burn_time) {
		struct timeval tv;

		tv.tv_sec  = 0;
		tv.tv_usec = burn_time;

		select(0, NULL, NULL, NULL, &tv);
	}
}

delay(x)
	volatile x;
{
	while (--x >= 0) {
		volatile int i = delay_time;
		while (--i >= 0);
	}
}

write_byte(addr, byte)
	unsigned addr, byte;
{
	ctl |= VPP;

	set_addr(addr + woff);

	outb(data, byte);
	delay(1);
	outb(control, ctl | ENB);
	burn_delay();
	outb(control, ctl);
}

read_byte(addr)
	unsigned addr;
{
	unsigned byte;

	ctl &= ~VPP;

	set_addr(addr + roff);

	outb(control, ctl | ENB | DIR);
	delay(1);
	byte = inb(data);
	outb(control, ctl);

	return (byte);
}

set_addr(addr)
	unsigned addr;
{
	addr += offset;

	outb(control, ctl);           delay(1);
	outb(data, addr >> 8);        delay(1);
	outb(control, ctl | ADDRESS); delay(1);
	outb(data, addr);             delay(1);
	outb(control, ctl);           delay(1);
}

